package %PACKAGE%; 

import java.util.ArrayList;

import android.content.SharedPreferences;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Point;
import android.graphics.PointF;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Typeface;
import android.os.Handler;
import android.preference.PreferenceCategory;
import android.preference.PreferenceManager;
import android.service.wallpaper.WallpaperService;
import android.util.DisplayMetrics;
import android.view.MotionEvent;
import android.view.SurfaceHolder;

public class SBLiveWallpaper extends WallpaperService {

    public static final String SHARED_PREFS_NAME="LiveWallpaperPrefs";
  
    @Override
    public void onCreate() {
        super.onCreate();
    }

    @Override
    public void onDestroy() {
        super.onDestroy();
    }

    @Override
    public Engine onCreateEngine() {
        return new LwpEngine();
    }

    class LwpEngine extends Engine implements SharedPreferences.OnSharedPreferenceChangeListener {

        private final Handler mHandler = new Handler();
        
        private float mTouchX = -1;
        private float mTouchY = -1;

        private final Runnable mDrawLWP = new Runnable() {
            public void run() {
                drawFrame();
            }
        };
        private boolean mVisible;
        private SharedPreferences mPrefs;
        
        // FANDO'S MODIFICATIONS
        
        private boolean mIsDemo;
        
        static final int FPS = 100;       
       	
    	Item[] mItems; // define an array which will hold our falling items   	
        
        // This is the correct way to get the visible screen dimensions: 
        // assign these variables in the onSurfaceChanged method
        private int mCanvasWidth = 0;
        private int mCanvasHeight = 0;
        
        private Resources mResources;
        private DisplayMetrics mDisplayMetrics;
        
        private int mItem_count = 0;
        
        private Bitmap mBackground;
        private Point mBackgroundOffset = new Point(0,0);
        private Bitmap[] mItemBitmaps;
        private Bitmap[] mPopImages;
               
        private int mPixelsX_offset; // how many pixels the screen has shifted (so we could draw the wallpaper parallax effect correctly)
       
        private String mSelectedSpeedValue;
        private int[] mSpeedValues;
        //private float[] mCalculatedSpeeds;
        
        // POINTS
        private boolean mShowStreakCounter = false;
        private int mStreakCounter = 0;
        private int mShowStreakEvery = 0;
        private boolean isStreak = false;
        
        // PAINT
        private Paint mPaint_PopText;
        private Paint mPaint_StreakCounter;
        private Paint mPaint_StreakPop;
        
        // POP objects
        private ArrayList<Pop_Item> mPopItems = new ArrayList<Pop_Item>();
        private String mPopStyle;        
        private String[] mPopText;
                
        private final int SPEED_VARIANCE = 18; // a slight variance of speed to further randomize item speed
        
        private boolean mDisableInteraction = false;
        private boolean mDisableInteractionChecked = false;
        private boolean mHideItemsChecked = false;
        
        private PointF mDirectionVector;
        private String mDirectionString;
        private final int POSITIONAL_VARIANCE = 3; // an interval spanning the indicated number of widths or heights when randomising screen item's reset position
        
        LwpEngine() {       
        	mIsDemo = getBooleanResource("isDemo");
        	int numOfItems = getIntegerResource("num_of_items");
        	
        	mItem_count = getStringArrayResourceCount("screen_items");
        	
        	mItems = new Item[mItem_count != 0 ? numOfItems : 0];
        	
            mPrefs = SBLiveWallpaper.this.getSharedPreferences(SHARED_PREFS_NAME, 0);
            mPrefs.registerOnSharedPreferenceChangeListener(this);
            onSharedPreferenceChanged(mPrefs, null);
            
            mResources = getResources();

           // init variables
            mDisplayMetrics = mResources.getDisplayMetrics();        
            
            mBackground = getWallpaperResource();
            //mBackground = BitmapFactory.decodeResource(mResources, R.drawable.wallpaper);
                    
            mItemBitmaps = getBitmaps("screen_items"); // create an array of all falling items      
            mPopImages = getBitmaps("action_items");  // create an array of all pop images      
            
            
            
            mPopText = getStringArrayResource("pop_text"); // load array of pop objects
            mPopStyle = getStringResource("pop_style");
            mShowStreakEvery = getIntegerResource("show_streak_every");
            
            mDisableInteraction = getBooleanResource("disableInteraction");
            
            /*
            mSelectedSpeedValue = mIsDemo ? "slow" : "standard";
            mSpeedValues = getIntegerArrayResource(mSelectedSpeedValue);
            
            mDirectionString = getStringResource("direction");
            mDirectionVector = getDirectionVector(mDirectionString);
            
            */
            
            initPaintObjects();            
        }  
        
        public void onSharedPreferenceChanged(SharedPreferences prefs, String key) { 
        	
        	
        	String counter_settings = "";
        	
        	if (mIsDemo) { 
        		mSelectedSpeedValue = "slow";  
        		//mDirectionString = "up";
        	}
        	else {
        		// hide items
        		mHideItemsChecked = prefs.getBoolean("hide_items", false);         		
        		
        		// interaction
        		mDisableInteractionChecked = prefs.getBoolean("disable_interaction", false); 
        		
        		// speed
        		mSelectedSpeedValue = prefs.getString("speed_settings", "standard"); 
	            
	            // counter
	            counter_settings = prefs.getString("counter_settings", "hide");	            
	            
	            
        	}
        	
            // direction
            mDirectionString = prefs.getString("direction_settings", getResources().getString(R.string.direction));            
            mSpeedValues = getIntegerArrayResource(mSelectedSpeedValue);      
            mShowStreakCounter = counter_settings.equals("show") && mItem_count > 0 && !mHideItemsChecked && !mDisableInteractionChecked;
            mDirectionVector = getDirectionVector(mDirectionString);
            
            // reset all items
            for (Item i : mItems) if (i != null) resetItem(i);
        }
        
        @Override
        public void onCreate(SurfaceHolder surfaceHolder) {
            super.onCreate(surfaceHolder);
            setTouchEventsEnabled(true);              
            PreferenceManager.setDefaultValues(getBaseContext(), R.xml.lwp_settings, false);
        }

        @Override
        public void onDestroy() {
            super.onDestroy();
            mHandler.removeCallbacks(mDrawLWP);
        }

        @Override
        public void onVisibilityChanged(boolean visible) {
            mVisible = visible;
            if (visible) {
                drawFrame();
            } else {
                mHandler.removeCallbacks(mDrawLWP);
            }
        }

        @Override
        public void onSurfaceChanged(SurfaceHolder holder, int format, int width, int height) {
            super.onSurfaceChanged(holder, format, width, height);            
            
            mCanvasWidth = width;
            mCanvasHeight = height;    
            
            
            initItems();   
            
            drawFrame();
        }

        @Override
        public void onSurfaceCreated(SurfaceHolder holder) {
            super.onSurfaceCreated(holder);
        }

        @Override
        public void onSurfaceDestroyed(SurfaceHolder holder) {
            super.onSurfaceDestroyed(holder);
            mVisible = false;
            mHandler.removeCallbacks(mDrawLWP);
        }

        @Override
        public void onOffsetsChanged(float xOffset, float yOffset, float xStep, float yStep, int xPixels, int yPixels) {
    		mPixelsX_offset = xPixels;
    		
    		//for (Item i : mItems)
    			//i.setPosition(i.getPos().x + xOffset * i.getWidth() * 0.5, i.getPos().y);
    		
            drawFrame();
        }

        
        @Override
        public void onTouchEvent(MotionEvent event) {   
        	if (!mDisableInteraction && !mDisableInteractionChecked) {
        	
	            mTouchX = event.getX();
	            mTouchY = event.getY();
            
	            if (event.getAction() == MotionEvent.ACTION_DOWN) {
	            	
	    			boolean isHit = false;
	    			PointF p;
	    			
	    	      	for (Item i : mItems) { 
	            		p = i.getPos(); // get item position   
	            		
	            		if (mTouchX >= p.x && mTouchX <= p.x + i.getWidth() && mTouchY >= p.y && mTouchY <= p.y + i.getHeight() && i.getState() == State.ALIVE) {
	            			mStreakCounter++;
	            			i.setDyingState(mStreakCounter);	
	            			isHit = true;		            			
	            			break;
	            		}
	    	      	}		    	      	
	    	      	
	    	      	if (!isHit) mStreakCounter = 0; 
	    	      	
	    		}
        	}
        	
            super.onTouchEvent(event);
        }

        void drawFrame() {
            final SurfaceHolder holder = getSurfaceHolder();

            Canvas c = null;
            try {
                c = holder.lockCanvas();
                if (c != null) {
                	update();
                	draw(c);
                }
            } finally {
                if (c != null) holder.unlockCanvasAndPost(c);
            }
            
            mHandler.removeCallbacks(mDrawLWP);
            if (mVisible) {
                mHandler.postDelayed(mDrawLWP, 1000 / FPS);
            }
        }
        
        void update(){   
        	Bitmap randomPopItem = null;
        	String randomPopText = "";
        	PointF pos;
        	PointF posItem = null;       	        	
    		
        	// update items
			for (Pop_Item p : mPopItems) {
				p.update(1000 / FPS);
			}
			
        	// remove dead items from array
			int p = 0;			
			while (p < mPopItems.size()) {
				if (!mPopItems.get(p).isAlive()) {
					mPopItems.remove(p);
				}
				else 
					p += 1;
			}			
        	
        	//for (Item i : mItems) {  
        	for (int k = mItems.length - 1; k >= 0; k--) {
        		Item i = mItems[k];
        		pos = i.getPos();
        		
        		if (!i.isAlive()) { // check if item was touched        			

        			if (mPopStyle.equals("image") || mPopStyle.equals("both")) {
        				randomPopItem = getRandomBitmap(mPopImages);
        				posItem = calcSpawnPosition(i, randomPopItem);
        			}
        			if (mPopStyle.equals("text") || mPopStyle.equals("both"))
        				randomPopText = getRandomText();	
        			//posText = calcSpawnPosition(i.getPos(), randomPopText);
        			
        			if (mShowStreakEvery == 0) isStreak = false;
    	      		else isStreak = (i.getStreak() % mShowStreakEvery == 0);
        			
        			if (isStreak) {
        				mPopItems.add(new Pop_Item("+" + i.getStreak(), calcSpawnPosition(i, "+" + i.getStreak(), mPaint_StreakPop), 600, mPaint_StreakPop));
        				isStreak = false;
        			}
        			else if (!mPopStyle.equals("none")) {
	        			if (mPopStyle.equals("text")) {
	        				mPopItems.add(new Pop_Item(randomPopText, calcSpawnPosition(i, randomPopText, mPaint_PopText), 300, mPaint_PopText));
	        			}
	        			else if (mPopStyle.equals("image")) {
	        				mPopItems.add(new Pop_Item(randomPopItem, posItem, 350));
	        			}
	        			else if (mPopStyle.equals("both")) {
	        				if (Math.random() < 0.5)
	        					mPopItems.add(new Pop_Item(randomPopText, calcSpawnPosition(i, randomPopText, mPaint_PopText), 300, mPaint_PopText));
	        				else 
	        					mPopItems.add(new Pop_Item(randomPopItem, posItem, 350));
	        			}
        			}
        			resetItem(i);
        		}        		

        		// move the item position down
        		i.update(1000 / FPS);     
        		
        		// check if item outside of bounds, if it is, reset it
        		if ( (mDirectionVector.x == 1 && pos.x > mCanvasWidth + POSITIONAL_VARIANCE * i.getWidth()) ||
        			 (mDirectionVector.x == -1 && pos.x < 0 - POSITIONAL_VARIANCE * i.getWidth()) || 
        			 (mDirectionVector.y == 1 && pos.y > mCanvasHeight + i.getHeight()) ||
        			 (mDirectionVector.y == -1 && pos.y < 0 - i.getHeight())) 
        				resetItem(i);
        	}
        	
        }
        
        void draw(Canvas c) {        	       	
        	// draw wallpaper background        	
        	c.drawBitmap(mBackground, mPixelsX_offset - mBackgroundOffset.x, 0 - mBackgroundOffset.y, null);   
        	
        	
        	// do not draw items whenever this options is set
        	if (!mHideItemsChecked) {        	
	        	// draw streak counter
	        	if (mShowStreakCounter) {
	        		float t = mPaint_StreakCounter.measureText(mStreakCounter + "");        		
	        		c.drawText(mStreakCounter + "", mCanvasWidth / 2 - t / 2, mCanvasHeight / 3, mPaint_StreakCounter);
	        	}
	        	
	        	        	
	        	// draw falling items        	
	        	//if (!mSelectedSpeedValue.equals("none")) {
		        	for (Item i : mItems) {  
		        		i.draw(c);
		        	}    
	        	//} 
	    		
	        	// draw poped items
				for (Pop_Item p : mPopItems) {
					p.draw(c);
				}   
        	}
        	
        }
        
        // Create an array of items which will be falling        
        void initItems() {       	 
        	
        	// for each item:	randomize x position (0 to width of screen - width of item) 
        	//					randomize display image
        	//					randomize speed
        	
        	Bitmap randomBitmap;
        	for (int i = 0; i < mItems.length; i++) {
        		randomBitmap = getRandomBitmap(mItemBitmaps);	
        		mItems[i] = new Item(getRandomPosition(randomBitmap), getRandomSpeed(), randomBitmap);
        	}
        }
        
        void resetItem(Item i){
        	Bitmap randomBitmap = getRandomBitmap(mItemBitmaps);			
        	i.setPosition(getRandomPosition(randomBitmap));// randomize X position and set Y position to top 		
			i.setImage(randomBitmap);// randomize image
			i.setSpeed(getRandomSpeed());// randomize speed
			i.setAliveState();
        }
        
        private Bitmap getRandomBitmap(Bitmap[] ar) {           	
        	return ar[randomInt(0, ar.length - 1)];
        }

        private PointF getRandomSpeed() {    
        	
    		int varMin = SPEED_VARIANCE / 2;
    		int randomSpeedValue = mSpeedValues[randomInt(0, mSpeedValues.length - 1)];
    		       		
    		int temp = randomInt(randomSpeedValue - varMin, randomSpeedValue + (int)((Math.random() > 0.5 ? -0.5 : 1) * SPEED_VARIANCE));
    		    
    		float speed = temp / 100f * mCanvasHeight / FPS;        		
    	
    		if (mDirectionVector.x == 0) return new PointF(0, mDirectionVector.y * speed); 
    		return new PointF(mDirectionVector.x * speed, 0); 
        }
        
        private String getRandomText() {        	
        	return mPopText[randomInt(0, mPopText.length - 1)];
        } 
        
        public PointF getRandomPosition(Bitmap b) {       	
        	
        	PointF pos = new PointF();
        	
        	if (mDirectionVector.x == 0) { 
        		pos.x = randomInt(0, mCanvasWidth - b.getWidth());
        		//pos.y = mDirectionVector.y * b.getHeight() + (mDirectionVector.y == -1 ? mCanvasHeight : 0);
        		pos.y = randomInt(b.getHeight(), POSITIONAL_VARIANCE * b.getHeight());
        		pos.y = mDirectionVector.y == 1 ? 0 - pos.y : mCanvasHeight + pos.y;       
        	}
        	else {
        	//if (mDirectionVector.y == 0) 
        	//pos.x = mDirectionVector.x * b.getWidth() + (mDirectionVector.x == -1 ? mCanvasWidth : 0);
	        	pos.x = randomInt(b.getWidth(), POSITIONAL_VARIANCE * b.getWidth());        	
	        	pos.x = mDirectionVector.x == 1 ? 0 - pos.x : mCanvasWidth + pos.x;        	
	        	pos.y = randomInt(0, mCanvasHeight - b.getHeight());
        	}
        	
         	return pos; 
        }       
        
        private PointF calcSpawnPosition(Item relativeTo, Bitmap b) {        	
        	
        	PointF p = relativeTo.getPos();
        	return new PointF(p.x + (relativeTo.getWidth() - b.getWidth()) / 2, p.y + (relativeTo.getHeight() - b.getHeight()) / 2);    
        }
        
        private PointF calcSpawnPosition(Item relativeTo, String text, Paint paint) {     
        	Rect rect = new Rect(0,0,0,0);
        	PointF p = relativeTo.getPos();
        	
        	paint.getTextBounds(text, 0, text.length(), rect);
            
            return new PointF(p.x + (relativeTo.getWidth() - rect.width()) / 2, p.y + (relativeTo.getHeight() - rect.height()) / 2);
        }
        
        public void initPaintObjects() {
        	int offset = 0;
        	
        	mPaint_PopText = new Paint();
        	mPaint_PopText.setAntiAlias(true);
        	mPaint_PopText.setColor(getColorResource("font_color_pop"));
        	mPaint_PopText.setTypeface(Typeface.createFromAsset(getAssets(), "fonts/" + getStringResource("font_name_pop") + ".ttf"));        	
        	if (getIntegerResource("font_size_pop") > 20) offset = 20;         	
        	mPaint_PopText.setTextSize(getIntegerResource("font_size_pop") - offset);
        	offset = 0;
        	//if (mPaint_PopText.getTextSize() > 20) mPaint_PopText.setTextSize(mPaint_PopText.getTextSize() - 20);
           	
        	mPaint_StreakCounter = new Paint();
        	mPaint_StreakCounter.setAntiAlias(true);
        	mPaint_StreakCounter.setColor(getColorResource("font_color_streak"));
        	mPaint_StreakCounter.setTypeface(Typeface.createFromAsset(getAssets(), "fonts/" + getStringResource("font_name_streak") + ".ttf"));
        	mPaint_StreakCounter.setTextSize((float)getIntegerResource("font_size_streak"));
        	//if (mPaint_PopText.getTextSize() > 20) mPaint_PopText.setTextSize(mPaint_PopText.getTextSize() - 20);
                 	
        	mPaint_StreakPop = new Paint();
        	mPaint_StreakPop.setAntiAlias(true);
        	mPaint_StreakPop.setColor(getColorResource("font_color_streakPop"));
        	mPaint_StreakPop.setTypeface(Typeface.createFromAsset(getAssets(), "fonts/" + getStringResource("font_name_streakPop") + ".ttf"));
        	//mPaint_StreakPop.setTextSize((float)getIntegerResource("font_size_streakPop"));
        	if (getIntegerResource("font_size_streakPop") > 20) offset = 20;         	
        	mPaint_StreakPop.setTextSize(getIntegerResource("font_size_streakPop") - offset);
        	//if (mPaint_StreakPop.getTextSize() > 20) mPaint_StreakPop.setTextSize(mPaint_StreakPop.getTextSize() - 20);
        }
        
        public Bitmap getWallpaperResource() {    	
        	
        	Bitmap bg;
        	Point wlp = new Point(1,1); // dimentions of wallpaper resource
        	DisplayMetrics dm = getResources().getDisplayMetrics();
        	int tWidth = dm.widthPixels * 2; // target width and height
        	int tHeight = dm.heightPixels;    	
        	
        	
        	// calculate target ratio
        	float tRatio = tWidth / (float)tHeight;    	
        	
        	
        	// get resource which best suits the target ratio
        	final float RATIO_HIGH = 1080f / 960;
        	final float RATIO_MID = 640f / 480;    			
        	
        	if (Math.abs(RATIO_HIGH - tRatio) < Math.abs(RATIO_MID - tRatio)) { // use high-dpi resource    		
        		bg = BitmapFactory.decodeResource(getResources(), R.drawable.wallpaper_hdpi); 
        		wlp.x = 1080;
        		wlp.y = 960;
        	}
        	else { // use mid-dpi resource
        		bg = BitmapFactory.decodeResource(getResources(), R.drawable.wallpaper_mdpi);   		
        		wlp.x = 640;
        		wlp.y = 480;    		
        	}        	
        	
        	// calculate scale factor
        	float scaleFactor = (float)tWidth / wlp.x; // assume horizontal constraint    	
        	if (wlp.y * scaleFactor < tHeight) scaleFactor = tHeight / (float)wlp.y; // attempt to scale, switch constraint to vertical if scaled height is smaller than target
        	
        	
        	// calculate new wallpaper dimensions
        	int scaledWidth = (int)(wlp.x * scaleFactor + 0.5);
        	int scaledHeight = (int)(wlp.y * scaleFactor + 0.5);
        	
        	// set x and y offsets (because the resulting scaled image may be larger than the given screen size we want to center it during drawing)
        	mBackgroundOffset = new Point(Math.abs(tWidth - scaledWidth) / 2, Math.abs(tHeight - scaledHeight) / 2);
        	
        	// scale wallpaper resource
        	return Bitmap.createScaledBitmap(bg, scaledWidth, scaledHeight, true); 	        
        }
    }
    
    public PointF getDirectionVector(String dir) {
    	int x = 1;
    	int y = 0;    	
    	if (dir.equals("left")) {x = -1; y = 0;}
    	if (dir.equals("right")) {x = 1; y = 0;}
    	if (dir.equals("up")) {x = 0; y = -1;}
    	if (dir.equals("down")) {x = 0; y = 1;}   	
    	
    	return new PointF(x, y);
    }
    
    // helper methods for retrieving resource values from xml
    public String getStringResource(String name) {    	
    	String str = ""; 
    	
    	try {
    		str = getResources().getString(getResources().getIdentifier(name, "string", getPackageName()));
    	}
    	catch(Exception ex) {    		
    	}    
    	
    	return str;
    }
    
    public boolean getBooleanResource(String name) {      	
    	boolean bl = false;
    	
    	try {
    		bl = getStringResource(name).equals("true") ? true : false;
    	}
    	catch(Exception ex) {}
    	
    	return bl;   	
    }
    
    public int getIntegerResource(String name) {    	
    	int integer = 0; 
    	
    	try {
    		integer = getResources().getInteger(getResources().getIdentifier(name, "integer", getPackageName()));
    	}
    	catch(Exception ex) {}    
    	
    	return integer;
    }
    
    public int getColorResource(String name) {   
    	int color = 0;
    	
    	try {
    		color = getResources().getColor(getResources().getIdentifier(name, "color", getPackageName()));
    	}
    	catch(Exception ex) {}
    	
    	return color;
    }
    
    public String[] getStringArrayResource(String name) {    	
    	String[] str = new String[0];
    	
    	try {
    		str = getResources().getStringArray(getResources().getIdentifier(name, "array", getPackageName()));
    	}
    	catch(Exception ex) {}
    	
    	return str;
	}
    
    public int[] getIntegerArrayResource(String name) {    	
    	int[] integers = new int[0];
    	
    	try {
    		integers = getResources().getIntArray(getResources().getIdentifier(name, "array", getPackageName()));
    	}
    	catch(Exception ex) {}
    	
    	return integers;  
	}
    
    // returns a random integer between min and max inclusive
    public static int randomInt(int min, int max) {    	
    	return min + (int)(Math.random() * (max - min + 1));
	}
    
    public static int randomInt(float min, float max) {    	
    	return (int)min + (int)(Math.random() * ((int)max - (int)min + 1));
	}
    
    public Bitmap[] getBitmaps(String key) {
    	Bitmap[] bitmaps;
    	
    	String[] names = getStringArrayResource(key);
    
    	if (names.length == 0) return new Bitmap[0];
    	
    	bitmaps = new Bitmap[names.length];
    	
    	for (int i = 0; i < names.length; i++)
    		bitmaps[i] = BitmapFactory.decodeResource(getResources(), getResources().getIdentifier(names[i], "drawable", getPackageName()));
    	
		return bitmaps;
    }
    
    public Bitmap[] getBitmapsArray(String prefix_key, String count_key) {
        String itemPrefix  = getStringResource(prefix_key);
        int itemCount  = getIntegerResource(count_key);

        Bitmap[] bitmaps = new Bitmap[itemCount];
        
        for (int i = 0; i < itemCount; i++) {
        	bitmaps[i] = BitmapFactory.decodeResource(getResources(), getResources().getIdentifier(itemPrefix + (i + 1), "drawable", getPackageName()));
        }
    	
        return bitmaps;
    }
    
    public int getStringArrayResourceCount(String key) {    	
    	return getStringArrayResource(key).length;
    }
    
    
}













